/*
 * Copyright (c) 2001, DecisionSoft Limited All rights reserved.
 * Please see LICENSE.TXT for more information.
 */

#include "../config/pathan_config.h"
#include <pathan/internal/ProxyMemoryManager.hpp>

#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/XMLException.hpp>
#include <xercesc/util/RuntimeException.hpp>
#include <xercesc/util/HashPtr.hpp>
#include <xercesc/dom/DOMNode.hpp>
#include <xercesc/dom/DOMDocument.hpp>

#include <pathan/internal/StringPool.hpp>
#include <pathan/PathanException.hpp>
#include <pathan/internal/context/XPath2ContextImpl.hpp>
#include <pathan/internal/DOMutils/XStr.hpp>
#include <pathan/internal/collations/CollationImpl.hpp>
#include <pathan/internal/dom-extensions/PathanNSResolverImpl.hpp>
#include "../simpleVariables/VarStoreImpl.hpp"
#include "../simpleVariables/VarTypeStoreImpl.hpp"
#include "../functionAPI/FunctionLookupImpl.hpp"
#include "../items/impl/ATDecimalOrDerivedImpl.hpp"
#include <pathan/functions/FunctionConstructor.hpp> // for datatype-uri

#include <iostream>

ProxyMemoryManager::ProxyMemoryManager(XERCES_CPP_NAMESPACE_QUALIFIER MemoryManager *parent)
  : parent_(parent)
{
  initialise();
}

ProxyMemoryManager::~ProxyMemoryManager() 
{
  releaseAll();
}

void ProxyMemoryManager::reset()
{
  releaseAll();
  initialise();
}

void ProxyMemoryManager::initialise()
{
  fCurrentBlock = 0;
  fStringPool = new (this) StringPool(this);
  fIntegerPool = new (this) XERCES_CPP_NAMESPACE_QUALIFIER RefHashTableOf<ATDecimalOrDerived>(53,true, new (this) XERCES_CPP_NAMESPACE_QUALIFIER HashPtr(),this);
}

void ProxyMemoryManager::releaseAll()
{
  fIntegerPool->cleanup();
  // Run backwards through the linked list, deleting the blocks of memory
  while(fCurrentBlock) {
    MemList *prev = fCurrentBlock->prev;
    parent_->deallocate((char *)fCurrentBlock);
    fCurrentBlock = prev;
  }
}

void *ProxyMemoryManager::allocate(size_t amount)
{
  char *newBlock = 0;
  try {
    newBlock = (char*)parent_->allocate(amount + sizeof(MemList));
  }
  catch (...) {}
  if(!newBlock) {
    //ThrowXML(RuntimeException, XMLExcepts::Out_Of_Memory);
    throw PathanException(PathanException::RUNTIME_ERR, X("ProxyMemoryManager::allocate(): Out of memory"));
  }

  MemList *newMemList = (MemList *)newBlock;
  newMemList->prev = fCurrentBlock;
  newMemList->next = 0;

  if(fCurrentBlock) {
    fCurrentBlock->next = newMemList;
  }
  fCurrentBlock = newMemList;

  return (void *)(newBlock + sizeof(MemList));
}

void ProxyMemoryManager::deallocate(void* p)
{
  if(p) {
    char *oldBlock = ((char *)p) - sizeof(MemList);
    MemList *oldMemList = (MemList *)oldBlock;

    // Decouple the old block from the list
    if(oldMemList->prev) {
      oldMemList->prev->next = oldMemList->next;
    }
    if(oldMemList->next) {
      oldMemList->next->prev = oldMemList->prev;
    }
    else {
      // If the old block doesn't have a next, it must currently be fCurrentBlock
      fCurrentBlock = oldMemList->prev;
    }

    parent_->deallocate(oldBlock);
  }
}

const XMLCh *ProxyMemoryManager::getPooledString(const XMLCh *src)
{
  return fStringPool->getPooledString(src);
}

const XMLCh *ProxyMemoryManager::getPooledString(const char *src)
{
  return fStringPool->getPooledString(src);
}

void ProxyMemoryManager::dumpStatistics()
{
  std::cout << "Memory statistics:" << std::endl;

  std::cout << "\tStrings in pool: " << fStringPool->getCount() << std::endl;

  MemList* pBlockCursor = fCurrentBlock;
  unsigned int nBlockCount=0;
  while(pBlockCursor) {
    ++nBlockCount;
    pBlockCursor = pBlockCursor->prev;
  }
  std::cout << "\tAllocated objects: " << nBlockCount << std::endl;
}

/** create a collation */
Collation* ProxyMemoryManager::createCollation(CollationHelper* helper) {
  return new (this) CollationImpl(this,helper);
}

/** create a resolver */
PathanNSResolver* ProxyMemoryManager::createNSResolver(XERCES_CPP_NAMESPACE_QUALIFIER DOMNode *resolverNode) {
  return new (this) PathanNSResolverImpl(this, resolverNode);
}

/** create a store for variables */
VariableStore* ProxyMemoryManager::createVariableStore() {
  return new (this) VarStoreImpl(this);
} 

/** create a store for variables */
VariableTypeStore* ProxyMemoryManager::createVariableTypeStore() {
  return new (this) VarTypeStoreImpl(this);
} 

/** create a function table */
FunctionLookup* ProxyMemoryManager::createFunctionTable() {
  return new (this) FunctionLookupImpl(this);
}

/** create a ATDecimalOrDerived for the given integer */
ATDecimalOrDerived* ProxyMemoryManager::createInteger(int value) {
  if(fIntegerPool->containsKey((const void*)value))
      return fIntegerPool->get((const void*)value);
  ATDecimalOrDerivedImpl* itemValue=new ATDecimalOrDerivedImpl(value);
  itemValue->incrementRefCount();
  fIntegerPool->put((void*)value, itemValue);
  return itemValue;
}
